@@ -572,8 +572,6 @@ class ContactService(BaseRdbmsService):
         :param connection: An object `RdbmsConnection` with auto commit.
 
         """
-        account = self.get_account(contact.account_id, check_status=True)
-
         with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
             cursor = connection.execute(
                 """
@@ -597,22 +595,6 @@ class ContactService(BaseRdbmsService):
 
             is_primary = cursor.fetch_one().get_value('is_primary')
 
-            # Enable the account of the user if it was pending.
-            if account.object_status == ObjectStatus.pending:
-                connection.execute(
-                    """
-                    UPDATE 
-                        account
-                      SET   
-                        object_status = %(OBJECT_STATUS_ENABLED)s
-                      WHERE
-                        account_id = %(account_id)s
-                    """,
-                    {
-                        'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
-                        'account_id': account_id
-                    })
-
             # Set this contact information as primary if the user doesn't have any
             # primary contact information.
             if not is_primary:
