@@ -25,10 +25,11 @@ import socket
 import urllib.request
 
 from majormode.perseus.constant.http import HttpMethod
+from majormode.perseus.constant.notification import DevicePlatform
 from majormode.perseus.constant.obj import ObjectStatus
 from majormode.perseus.constant.sort_order import SortOrder
 from majormode.perseus.model import obj
-from majormode.perseus.model.enum import Enum
+from majormode.perseus.model.locale import Locale
 from majormode.perseus.utils import cast
 from majormode.perseus.utils.date_util import ISO8601DateTime
 
@@ -90,25 +91,267 @@ class NotificationService(BaseRdbmsService):
     # to the target Android application running on the target device.
     GCM_SERVER_HTTP_URL = 'https://gcm-http.googleapis.com/gcm/send'
 
-    # Define the code name of the supported mobile device platforms.
-    DevicePlatform = Enum(
-        'ios',
-        'android'
-    )
+    def __add_device_registration(
+            self,
+            device_id,
+            device_platform,
+            app_id,
+            device_token,
+            account_id=None,
+            connection=None,
+            locale=None,
+            utc_offset=None):
+        """
+        Register a mobile device to the push notification service
 
-    NotificationMode = Enum(
-        # Indicate that a notification message is delivered to the specified
-        # recipients when the request for the transmission of information is
-        # initiated by the receiver or client application, and then is
-        # responded by the publisher or server platform.
-        'pull',
 
-        # Indicate that a notification message is delivered to the specified
-        # recipients when the request for the transmission of information is
-        # initiated by the publisher or server platform and pushed out to the
-        # receiver or client application.
-        'push'
-    )
+        :param device_id: Identification of a mobile device that registers
+            to push notification service.
+
+        :param device_platform: An item of the enumeration `DevicePlatform`.
+
+        :param app_id: Identification of the mobile application that
+            registers the mobile device to push notification service.
+
+        :param device_token: Token that identifies the device by the push
+            notification provider of the device platform.
+
+        :param account_id: Identification of the account of a user who
+            registers his mobile device to push notification service.
+
+        :param connection: An object `RdbmsConnection`.
+
+        :param locale: Preferred language to new messages' textual content in.
+            This argument is ignored if the mobile device is registered to
+            push notification on behalf of a user. Messages will be send in
+            the current preferred language of the user.
+
+        :param utc_offset: Difference between the location of the device and
+            UTC (Universal Time Coordinated).
+
+
+        :return: An object containing the following attributes:
+
+            * `device_id` (required): Identification of the mobile device that
+              registers to push notification service.
+
+            * `object_status` (required): Current status of the registration of
+              this device to push notification service.
+
+            * `registration_id` (required): Identification of the registration of
+              the mobile device to the push notification service.
+
+            * `update_time` (required): Time of the most recent modification of
+              information of the mobile device's registration to the push
+              notification service.
+        """
+        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
+            cursor = connection.execute("""
+                INSERT INTO notification_device (
+                    device_token,
+                    device_id,
+                    device_platform,
+                    account_id,
+                    utc_offset,
+                    locale,
+                    app_id)
+                  VALUES (
+                    %(device_token)s,
+                    %(device_id)s,
+                    %(device_platform)s,
+                    %(account_id)s,
+                    %(utc_offset)s,
+                    %(locale)s,
+                    %(app_id)s
+                  )
+                  RETURNING
+                    device_id,
+                    object_status,
+                    registration_id,
+                    update_time
+                """,
+                {
+                    'account_id': account_id,
+                    'app_id': app_id,
+                    'device_id': device_id,
+                    'device_platform': device_platform,
+                    'device_token': device_token,
+                    'locale': None if account_id else locale,
+                    'utc_offset': utc_offset
+                }
+            )
+
+            row = cursor.fetch_one()
+            device_registration = row.get_object({
+                'object_status': ObjectStatus,
+                'registration_id': cast.string_to_uuid,
+                'update_time': cast.string_to_timestamp,
+            })
+
+            return device_registration
+
+    def __get_device_registration(
+            self,
+            device_id,
+            app_id,
+            account_id=None,
+            connection=None):
+        """
+        Return the registration of a mobile device to push notification
+        service
+
+
+        :param device_id: Identification of a mobile device that registered
+            to push notification service.
+
+        :param app_id: Identification of the mobile application that
+            registered the mobile device to push notification service.
+
+        :param account_id: Identification of the account of a user who
+            registered his mobile device to push notification service.
+
+        :param connection: An object `RdbmsConnection`.
+
+
+        :return: An object containing the following attributes:
+
+            * `account_id` (optional): Identification of the account of the user
+              who registered his mobile device to push notification service.
+
+            * `app_id` (required): Identification of the mobile application that
+              registered the mobile device to push notification service.
+
+            * `device_id` (required): Identification of the mobile device that
+              registered to push notification service.
+
+            * `device_platform` (required): An item of the enumeration
+              `DevicePlatform` describing the platform of the mobile device.
+
+            * `device_token` (required): Token that identifies the mobile device
+              by the push notification provider of the device platform.
+
+            * `locale` (optional): Preferred language to receive new message's
+              textual content in.  This argument is not used if the device
+              registered to push notification service on behalf of a user, but the
+              preferred language of this user.
+
+            * `object_status` (required): Current status of the registration of
+              this device to push notification service
+
+            * `registration_id` (required): Identification of the registration of
+              the mobile device to the push notification service.
+
+            * `update_time` (required): Time of the most recent modification of
+              information of the mobile device's registration to the push
+              notification service.
+
+            * `utc_offset` (optional): Difference between the location of the
+               mobile device and UTC (Universal Time Coordinated).
+        """
+        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
+            cursor = connection.execute(
+                """
+                SELECT
+                    account_id,
+                    app_id,
+                    device_id,
+                    device_platform,
+                    device_token,
+                    locale,
+                    object_status,
+                    registration_id,
+                    update_time,
+                    utc_offset
+                  FROM 
+                    notification_device
+                  WHERE
+                    device_id = %(device_id)s
+                    AND ((account_id IS NULL AND %(account_id)s IS NULL)
+                         OR account_id = %(account_id)s)
+                    AND app_id = %(app_id)s
+                """,
+                {
+                    'account_id': account_id,
+                    'app_id': app_id,
+                    'device_id': device_id
+                })
+
+            row = cursor.fetch_one()
+
+            device_registration = row and row.get_object({
+                'account_id': cast.string_to_uuid,
+                'app_id': cast.string_to_uuid,
+                'device_platform': DevicePlatform,
+                'locale': Locale,
+                'object_status': ObjectStatus,
+                'registration_id': cast.string_to_uuid,
+                'update_time': cast.string_to_timestamp,
+            })
+
+            return device_registration
+
+    def __update_device_registration_token(
+            self,
+            registration_id,
+            device_token,
+            connection=None):
+        """
+        Update the token of a device registered to push notification service
+
+
+        :param registration_id: Identification of the registration of a device
+            to push notification service
+
+        :param device_token: Token that identifies the device by the push
+            notification provider of the device platform.
+
+        :param connection: An object `RdbmsConnection` with auto commit.
+
+
+        :return: An object containing the following attributes:
+
+            * `device_id` (required): Identification of the mobile device that
+              registers to push notification service.
+
+            * `object_status` (required): Current status of the registration of
+              this device to push notification service.
+
+            * `registration_id` (required): Identification of the registration of
+              the mobile device to the push notification service.
+
+            * `update_time` (required): Time of the most recent modification of
+              information of the mobile device's registration to the push
+              notification service.
+        """
+        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
+            cursor = connection.execute(
+                """
+                UPDATE 
+                    notification_device
+                  SET
+                    device_token = %(device_token)s,
+                    update_time = current_timestamp
+                  WHERE
+                    registration_id = %(registration_id)s
+                  RETURNING
+                    device_id,
+                    object_status,
+                    registration_id,
+                    update_time
+                """,
+                {
+                    'registration_id': registration_id,
+                    'device_token': device_token
+                })
+
+            row = cursor.fetch_one()
+            device_registration = row.get_object({
+                'object_status': ObjectStatus,
+                'registration_id': cast.string_to_uuid,
+                'update_time': cast.string_to_timestamp,
+            })
+
+            return device_registration
 
     def _push_android_notification(
             self,
@@ -174,7 +417,7 @@ class NotificationService(BaseRdbmsService):
             * `canonical_ids:integer` (required): number of results that contain
               a canonical registration token.
 
-            * `results:list` (optiona): array of objects representing the status
+            * `results:list` (optional): array of objects representing the status
               of the messages processed. The objects are listed in the same order
               as the request (i.e., for each registration ID in the request, its
               result is listed in the same index in the response):
@@ -312,90 +555,91 @@ class NotificationService(BaseRdbmsService):
             self,
             app_id,
             recipient_id,
-            start_time=None,
+            connection=None,
             end_time=None,
-            notification_types=None,
-            offset=0,
-            limit=BaseRdbmsService.DEFAULT_LIMIT,
             include_read=False,
+            limit=BaseRdbmsService.DEFAULT_LIMIT,
             mark_read=True,
-            sort_order=SortOrder.ascending):
+            offset=0,
+            notification_types=None,
+            sort_order=SortOrder.ascending,
+            start_time=None):
         """
         Return a list of notifications that have been sent to the specified
         recipient.
 
 
-        :param app_id: identification of the client application such as a Web,
+        :param app_id: Identification of the client application such as a Web,
             a desktop, or a mobile application, that accesses the service.
 
-        :param recipient_id: identification of a recipient that may have been
+        :param recipient_id: Identification of a recipient that may have been
             issued notifications, such as, for instance the identification of
             a user account, or the identification of a device, such as an
             International Mobile Equipment Identity (IMEI) for Android device,
             or an alphanumeric string for iOS device.
 
-        :param start_time: indicate the earliest time of submission to return
+        :param connection: An object `RdbmsConnection`.  Auto commit MUST be
+            enabled if the argument `mark_read` is equal to `True`.
+
+        :param end_time: Indicate the latest time of submission to return
             notifications.
 
-        :param end_time: indicate the latest time of submission to return
-            notifications.
+        :param include_read: Indicate whether to include notifications that
+            have been already read.
 
-        :param notification_types: a list of notification types the recipient
+        :param limit: Constrain the number of notifications that are returned
+            to the specified number.  Default value is
+            `NotificationService.DEFAULT_LIMIT`.  Maximum value is
+            is `NotificationService.MAXIMUM_LIMIT`.
+
+        :param mark_read: Indicate whether to mark as read every notification
+            that are returned.
+
+        :param notification_types: A list of notification types the recipient
             is interested in, or whatever notification if not defined.
 
-        :param offset: require to skip that many records before beginning to
+        :param offset: Require to skip that many records before beginning to
             return notifications.  Default value is `0`.  If both `offset`
             and `limit` are specified, then `offset` records are skipped
             before starting to count the limit notifications that are
             returned.
 
-        :param limit: constrain the number of notifications that are returned
-            to the specified number.  Default value is
-            `NotificationService.DEFAULT_LIMIT`.  Maximum value is
-            is `NotificationService.MAXIMUM_LIMIT`.
-
-        :param include_read: indicate whether to include notifications that
-            have been already read.
-
-        :param mark_read: indicate whether to mark as read every notification
-            that are returned.
-
-        :param sort_order: ascending order sorts notifications by ascending
+        :param sort_order: Ascending order sorts notifications by ascending
             schedule time, while descending order sorts notifications by
             descending schedule time.
 
+        :param start_time: Indicate the earliest time of submission to return
+            notifications.
 
-        :return: a list of instances containing the following members:
 
-            * `creation_time` (required): time when the sender originated
-              the notification to the intended recipient.
+        :return: A list of instances containing the following members:
 
-            * `is_read` (required): indicate whether the notification
-               has been read by the intended recipient.
+            * `creation_time` (required): Time when the sender originated the
+              notification to the intended recipient.
 
-            * `notification_id` (required): identification of the
-              notification.
+            * `is_read` (required): Indicate whether the notification has been read
+              by the intended recipient.
 
-            * `notification_type` (required): string representation of the
-              type of the notification, as selected by the sender that
-              originated this notification to the intended recipient.
+            * `notification_id` (required): Identification of the  notification.
 
-            * `payload` (optional): an arbitrary JSON expression added by
-              the sender to provide information about the context of this
-              notification.
+            * `notification_type` (required): String representation of the type of
+              the notification, as selected by the sender that originated this
+              notification to the intended recipient.
 
-            * `schedule_time` (required): time when this notification is
-              scheduled to be sent to the intended recipient.
-              The notification is not visible to the intended recipient prior
-              to this time.
+            * `payload` (optional): An arbitrary JSON expression added by the
+              sender to provide information about the context of this notification.
 
-            * `sender_id` (optional): the identification of the sender that
+            * `schedule_time` (required): Time when this notification is scheduled
+              to be sent to the intended recipient.  The notification is not visible
+              to the intended recipient prior to this time.
+
+            * `sender_id` (optional): The identification of the sender that
               originated the notification.
 
-            * `update_time` (required): time of the most recent modification
-              of an attribute of the notification, such as its read status.
+            * `update_time` (required): Time of the most recent modification of an
+              attribute of the notification, such as its read status.
         """
-        with self.acquire_rdbms_connection(True) as connection:
+        with self.acquire_rdbms_connection(auto_commit=mark_read, connection=connection) as connection:
             if mark_read:
                 cursor = connection.execute(
                     f"""
@@ -414,10 +658,10 @@ class NotificationService(BaseRdbmsService):
                               AND (%(start_time)s IS NULL OR creation_time > %(start_time)s)
                               AND (%(end_time)s IS NULL OR creation_time < %(end_time)s)
                               AND (%(include_read)s OR NOT is_read)
-                              AND (%(notification_types)s IS NULL OR notification_type IN (%[notification_types]s))
+                              AND (%(notification_types)s IS NULL OR notification_type IN (%(notification_types)s))
                               AND object_status = %(OBJECT_STATUS_ENABLED)s
                             ORDER BY
-                             creation_time {'ASC' if sort_order == SortOrder.ascending else 'DESC'}
+                              creation_time {'ASC' if sort_order == SortOrder.ascending else 'DESC'}
                             OFFSET %(offset)s
                             LIMIT %(limit)s)
                       RETURNING 
@@ -434,7 +678,7 @@ class NotificationService(BaseRdbmsService):
                         'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                         'end_time': end_time,
                         'include_read': include_read,
-                        'limit': min(limit and self.DEFAULT_LIMIT, self.MAXIMUM_LIMIT),
+                        'limit': min(limit, self.MAXIMUM_LIMIT) or self.DEFAULT_LIMIT,
                         'notification_types': notification_types,
                         'offset': offset,
                         'recipient_id': recipient_id,
@@ -460,7 +704,7 @@ class NotificationService(BaseRdbmsService):
                         AND (%(start_time)s IS NULL OR creation_time > %(start_time)s)
                         AND (%(end_time)s IS NULL OR creation_time < %(end_time)s)
                         AND (%(include_read)s OR NOT is_read)
-                        AND (%(notification_types)s IS NULL OR notification_type IN (%[notification_types]s))
+                        AND (%(notification_types)s IS NULL OR notification_type IN (%(notification_types)s))
                         AND object_status = %(OBJECT_STATUS_ENABLED)s
                       ORDER BY
                         creation_time {'ASC' if sort_order == SortOrder.ascending else 'DESC'}
@@ -471,7 +715,7 @@ class NotificationService(BaseRdbmsService):
                         'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                         'end_time': end_time,
                         'include_read': include_read,
-                        'limit': min(limit and self.DEFAULT_LIMIT, self.MAXIMUM_LIMIT),
+                        'limit': min(limit, self.MAXIMUM_LIMIT) or self.DEFAULT_LIMIT,
                         'notification_types': notification_types,
                         'offset': offset,
                         'recipient_id': recipient_id,
@@ -498,6 +742,7 @@ class NotificationService(BaseRdbmsService):
             device_token,
             device_platform,
             account_id=None,
+            connection=None,
             locale=None,
             topics=None,
             utc_offset=None):
@@ -518,10 +763,10 @@ class NotificationService(BaseRdbmsService):
         @todo: the argument "topics" is not used at the moment.
 
 
-        :param app_id: identification of the client application such as a Web,
+        :param app_id: Identification of the client application such as a Web,
             a desktop, or a mobile application, that accesses the service.
 
-        :param device_id: identification of the device, which depends on the
+        :param device_id: Identification of the device, which depends on the
             device platform:
 
             * Android: International Mobile Equipment Identity (IMEI) number
@@ -534,7 +779,7 @@ class NotificationService(BaseRdbmsService):
               solution aimed at solving the main privacy issues that caused
               Apple to deprecate UDIDs.
 
-        :param device_token: token that identifies the device by the push
+        :param device_token: Token that identifies the device by the push
             notification provider of the device platform.
 
             * Android: token identifying the device to push the notification
@@ -559,7 +804,7 @@ class NotificationService(BaseRdbmsService):
               as the device UDID returned by the `uniqueIdentifier` property
               of `UIDevice`.
 
-        :param device_platform: indicate the platform of the end user's mobile
+        :param device_platform: Indicate the platform of the end user's mobile
             device:
 
             * `ios`: Apple iOS
@@ -568,11 +813,13 @@ class NotificationService(BaseRdbmsService):
 
             * `windows`: Windows Phone
 
-        :param account_id: identification of the account of the user on behalf
+        :param account_id: Identification of the account of the user on behalf
             of whom the device is registered to receive push notification
             messages.
 
-        :param locale: represent the language that the end user prefers
+        :param connection: An object `RdbmsConnection` with auto commit.
+
+        :param locale: Represent the language that the end user prefers
             receiving new content in.  A locale corresponds to a tag
             respecting RFC 4646, expressed by a ISO 639-3 alpha-3 code
             element, optionally followed by a dash character `-` and a ISO
@@ -581,92 +828,60 @@ class NotificationService(BaseRdbmsService):
             argument is not specified, the locale corresponds to the preferred
             language that the user has defined in his profile.
 
-        :param topics: a list of keywords representing topics the end user is
+        :param topics: A list of keywords representing topics the end user is
             interested in to be pushed new content whenever related to one of
             those topics.  The list of supported keywords is specific to the
             publisher service of the client application and as such the
             developer of the client application has to refer to the technical
             documentation of the publisher service.
 
-        :param utc_offset: difference between the location of the device and
+        :param utc_offset: Difference between the location of the device and
             UTC (Universal Time Coordinated).  UTC is also known as GMT or
             Greenwich Mean Time or Zulu Time.
+
+
+        :return: An object containing the following attributes:
+
+            * `device_id` (required): Identification of the mobile device that
+              registers to push notification service.
+
+            * `object_status` (required): Current status of the registration of
+              this device to push notification service.
+
+            * `registration_id` (required): Identification of the registration of
+              the mobile device to the push notification service.
+
+            * `update_time` (required): Time of the most recent modification of
+              information of the mobile device's registration to the push
+              notification service.
         """
-        application = ApplicationService().get_application(app_id, check_status=True)
 
-        with self.acquire_rdbms_connection(True) as connection:
-            cursor = connection.execute(
-                """
-                SELECT 
-                    device_token
-                  FROM 
-                    notification_device
-                  WHERE
-                    device_id = %(device_id)s
-                    AND ((account_id IS NULL AND %(account_id)s IS NULL) OR
-                         (account_id = %(account_id)s))
-                    AND app_id = %(app_id)s
-                """,
-                {
-                    'account_id': account_id,
-                    'app_id': app_id,
-                    'device_id': device_id
-                })
+        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
+            ApplicationService().get_application(app_id, check_status=True, connection=connection)
 
-            row = cursor.fetch_one()
+            device_registration = self.__get_device_registration(
+                device_id,
+                app_id,
+                account_id=account_id,
+                connection=connection)
 
-            if row is None:
-                connection.execute("""
-                    INSERT INTO notification_device(
-                        device_token,
-                        device_id,
-                        device_platform,
-                        account_id,
-                        utc_offset,
-                        locale,
-                        app_id)
-                      VALUES (
-                        %(device_token)s,
-                        %(device_id)s,
-                        %(device_platform)s,
-                        %(account_id)s,
-                        %(utc_offset)s,
-                        %(locale)s,
-                        %(app_id)s)
-                    """,
-                    {
-                        'account_id': account_id,
-                        'app_id': app_id,
-                        'device_id': device_id,
-                        'device_platform': device_platform,
-                        'device_token': device_token,
-                        'locale': locale,
-                        #'package': application.package_name,
-                        'utc_offset': utc_offset
-                    })
-            else:
-                notification_device = row.get_object()
+            if device_registration is None:
+                device_registration = self.__add_device_registration(
+                    device_id,
+                    device_platform,
+                    app_id,
+                    device_token,
+                    account_id=account_id,
+                    connection=connection,
+                    locale=locale,
+                    utc_offset=utc_offset)
+            elif device_registration.device_token != device_token:
+                device_registration = self.__update_device_registration_token(
+                    device_registration.registration_id,
+                    device_token,
+                    connection=connection)
 
-                if notification_device.device_token != device_token:
-                    connection.execute(
-                        """
-                        UPDATE 
-                            notification_device
-                          SET
-                            device_token = %(device_token)s,
-                            update_time = current_timestamp
-                          WHERE
-                            device_id = %(device_id)s
-                            AND ((account_id IS NULL AND %(account_id)s IS NULL) OR
-                                 (account_id = %(account_id)s))
-                            AND app_id = %(app_id)s
-                        """,
-                        {
-                            'account_id': account_id,
-                            'app_id': app_id,
-                            'device_id': device_id,
-                            'device_token': device_token
-                        })
+            return device_registration
 
     def send_notification(
             self,
@@ -794,7 +1009,6 @@ class NotificationService(BaseRdbmsService):
             while True:
                 accounts = TeamService().get_members(
                     app_id,
-                    settings.PLATFORM_BOTNET_ACCOUNT_ID,
                     team_id,
                     limit=TeamService.DEFAULT_LIMIT,
                     offset=offset)
